from datetime import datetime, timedelta

from sqlalchemy import and_

from .models import session_ctx_manager, Task


def dashboard_cleaning(status, threshold):
    tasks_table = Task.__table__
    stm = (tasks_table
           .delete()
           .where(and_(tasks_table.c.status == status,
                       tasks_table.c.date_done <= datetime.utcnow() - timedelta(seconds=threshold))))
    with session_ctx_manager() as session:
        session.execute(stm)
        session.commit()



from celery.bin.base import Command

from .api.app import get_app


class CeleryDashboard(Command):
    def add_arguments(self, parser):
        parser.add_argument(
            '--port', help='Port to use', type=int
        )

    def run(self, port=5000, **kwargs):
        flask_app = get_app(self.app)
        flask_app.run(host="0.0.0.0", port=port)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
from flask import Flask, send_from_directory
from flask_sqlalchemy import SQLAlchemy

from ..api.static import static
from ..api.api import api


def get_app(celery_app):
    app = Flask(__name__, template_folder="frontend/dist")

    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = celery_app.conf.dashboard_pg_uri

    app.db = SQLAlchemy(app)
    app.celery_app = celery_app

    app.register_blueprint(api)
    app.register_blueprint(static)

    @app.route("/")
    def index():
        dist_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), "frontend/dist")
        return send_from_directory(dist_directory, "index.html")

    return app



from functools import wraps
from flask import request, Response, current_app


def check_auth(username, password):
    conf = current_app.celery_app.conf
    return username == conf.dashboard_username and password == conf.dashboard_password


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response(
        'Could not verify your access level for that URL.\n'
        'You have to login with proper credentials', 401,
        {'WWW-Authenticate': 'Basic realm="Login Required"'})


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.authorization
        conf = current_app.celery_app.conf
        if conf.dashboard_username and conf.dashboard_password:
            if not auth or not check_auth(auth.username, auth.password):
                return authenticate()
        return f(*args, **kwargs)

    return decorated



from datetime import datetime, timedelta

import pytz
from celery import current_app
from celery.signals import before_task_publish, task_prerun, task_retry, task_success, task_failure, task_revoked

from .models import Task
from .utils import dump


def check_restricted_statuses(status, task_name_getter):
    def decorator(receiver):
        def wrapper(sender=None, **kwargs):
            if hasattr(current_app.tasks[task_name_getter(sender)], "only_store"):
                if status not in current_app.tasks[task_name_getter(sender)].only_store:
                    return
            return receiver(sender, **kwargs)

        return wrapper

    return decorator


@before_task_publish.connect
@check_restricted_statuses(status="QUEUED", task_name_getter=lambda x: x)
def task_sent_handler(sender=None, headers=None, body=None, properties=None, **kwargs):
    # information about task are located in headers for task messages
    # using the task protocol version 2.
    info = (headers if headers and 'task' in headers else body) or {}
    argsrepr = info.get("argsrepr") or dump(info["args"])
    kwargsrepr = info.get("kwargsrepr") or dump(info["kwargs"])
    now = pytz.UTC.localize(datetime.utcnow())
    eta = info.get("eta") or now
    Task.upsert(info["id"], status="QUEUED", date_queued=now, name=sender,
                routing_key=kwargs.get("routing_key"),
                args=argsrepr, kwargs=kwargsrepr, on_conflict_update={"date_queued": now}, eta=eta)


@task_prerun.connect
@check_restricted_statuses(status="STARTED", task_name_getter=lambda x: x.name)
def task_started_handler(sender=None, task_id=None, args=None, kwargs=None, **opts):
    now = pytz.UTC.localize(datetime.utcnow())
    Task.upsert(task_id, status="STARTED", name=sender.name, args=dump(sender.request.args),
                kwargs=dump(sender.request.kwargs), date_started=now,
                routing_key=sender.request.delivery_info["routing_key"])


@task_retry.connect
@check_restricted_statuses(status="RETRY", task_name_getter=lambda x: x.name)
def task_retry_handler(sender=None, reason=None, request=None, einfo=None, **opts):
    when = getattr(getattr(einfo, "exception", None), "when")
    eta = None
    if isinstance(when, datetime):
        eta = when
    elif isinstance(when, int):
        eta = pytz.UTC.localize(datetime.utcnow()) + timedelta(seconds=when)
    Task.upsert(request.id, status="RETRY", name=sender.name, routing_key=request.delivery_info["routing_key"],
                exception_type=str(reason), args=dump(sender.request.args), kwargs=dump(sender.request.kwargs),
                traceback=str(einfo),
                date_done=pytz.UTC.localize(datetime.utcnow()), eta=eta)


@task_success.connect
@check_restricted_statuses(status="SUCCESS", task_name_getter=lambda x: x.name)
def task_success_handler(sender=None, result=None, **opts):
    resultrepr = dump(result)
    Task.upsert(sender.request.id, status="SUCCESS", name=sender.name,
                routing_key=sender.request.delivery_info["routing_key"],
                result=resultrepr, args=dump(sender.request.args), kwargs=dump(sender.request.kwargs),
                date_done=pytz.UTC.localize(datetime.utcnow()))


@task_failure.connect
@check_restricted_statuses(status="FAILURE", task_name_getter=lambda x: x.name)
def task_failure_handler(sender=None, exception=None, einfo=None, **opts):
    Task.upsert(sender.request.id, status="FAILURE", name=sender.name,
                routing_key=sender.request.delivery_info["routing_key"],
                exception_type=einfo.type.__name__, traceback=str(einfo.traceback), args=dump(sender.request.args),
                kwargs=dump(sender.request.kwargs),
                date_done=pytz.UTC.localize(datetime.utcnow()))


@task_revoked.connect
@check_restricted_statuses(status="REVOKED", task_name_getter=lambda x: x.name)
def task_revoked_handler(sender=None, request=None, **opts):
    argsrepr = getattr(request, "argsrepr", None) or dump(request.args)
    kwargsrepr = getattr(request, "kwargsrepr", None) or dump(request.kwargs)
    Task.upsert(request.id, status="REVOKED", name=sender.name,
                routing_key=request.delivery_info["routing_key"],
                args=argsrepr, kwargs=kwargsrepr,
                date_done=pytz.UTC.localize(datetime.utcnow()))



from multiprocessing.util import register_after_fork

from sqlalchemy import create_engine

from .cleaning import dashboard_cleaning
from .models import SessionMaker, prepare_models


def init(celery_app, pg_db_uri, username=None, password=None, cleaning_thresholds=None, db_echo=False):
    # setup db
    db_engine = create_engine(pg_db_uri, client_encoding='utf8', convert_unicode=True, echo=db_echo)
    SessionMaker.configure(bind=db_engine)
    prepared = prepare_models(db_engine)
    if not prepared:
        return

    from .signals import task_sent_handler, task_started_handler, task_retry_handler, task_success_handler, \
        task_failure_handler, task_revoked_handler
    register_after_fork(db_engine, lambda engine: engine.dispose())

    # cleaning
    if not cleaning_thresholds:
        cleaning_thresholds = {}
    # if task hasn't finished after 3600s assume worker has been killed and couldn't send a signal
    if "STARTED" not in cleaning_thresholds:
        cleaning_thresholds["STARTED"] = 3600
    if "SUCCESS" not in cleaning_thresholds:
        cleaning_thresholds["SUCCESS"] = 3600 * 4
    celery_app.conf.dashboard_pg_uri = pg_db_uri
    celery_app.conf.dashboard_username = username
    celery_app.conf.dashboard_password = password
    celery_app.task(name="dashboard_cleaning")(dashboard_cleaning)

    from celery import __version__ as celery_version
    for status, threshold in cleaning_thresholds.items():
        if celery_version.startswith('4'):
            beat_schedule_name = "beat_schedule"
        else:
            beat_schedule_name = "CELERYBEAT_SCHEDULE"
        getattr(celery_app.conf, beat_schedule_name)['clean-%s-tasks' % status.lower()] = {
            'task': 'dashboard_cleaning',
            'schedule': threshold,
            'args': (status, threshold),
            'options': {
                'queue': 'celery_dashboard',
                'expires': 10 * 60
            }
        }



import json

from flask import Blueprint, abort
from flask import current_app
from flask import request
from sqlalchemy import func

from ..auth import requires_auth
from ..models import Task
from ..utils import cancel_tasks, requeue_tasks

api = Blueprint('api', __name__, url_prefix='/api')


def with_task(func):
    def decorator(task_id, *args, **kwargs):
        task = current_app.db.session.query(Task).filter(Task.task_id == task_id).one()
        return func(task, *args, **kwargs)

    decorator.__name__ = func.__name__
    return decorator


@api.route("/tasks", methods=["GET", "DELETE", "POST"])
@requires_auth
def tasks():
    q = current_app.db.session.query(Task)

    if request.method == "GET":
        data = request.args
    else:
        data = request.get_json()

    if "status" in data:
        q = q.filter(Task.status == data["status"])

    if "queue" in data:
        q = q.filter(Task.routing_key == data["queue"])

    if "task" in data:
        q = q.filter(Task.name == data["task"])

    if "exception" in data:
        q = q.filter(Task.exception_type == data["exception"])

    if "taskId" in data:
        q = q.filter(Task.task_id == data["taskId"])

    if request.method == "GET":
        count = q.count()
        # add sorting here
        sorts = data.get("sort") or ""
        if sorts:
            for sort_str in sorts.split(","):
                col, direction = sort_str.split(":")
                q = q.order_by(getattr(getattr(Task, col), direction)())

        # offset and limit
        offset = int(request.args.get("start") or 0)
        limit = int(request.args.get("end") or 5) - offset
        q = q.order_by(Task.date_queued.asc()).offset(offset).limit(limit)

        result = [task.serialized for task in q.yield_per(1000)]

        return json.dumps({"result": result, "count": count})
    elif request.method == "DELETE":
        count = cancel_tasks(q.yield_per(1000), current_app.db.session)
        current_app.db.session.commit()
        return json.dumps({"count": count})
    elif request.method == "POST":
        task_ids = requeue_tasks(q.yield_per(1000), current_app.db.session)
        current_app.db.session.commit()
        return json.dumps({"count": len(task_ids)})


@api.route("/task/<task_id>/revoke")
@requires_auth
@with_task
def revoke_task(task):
    cancel_tasks([task], current_app.db.session)
    current_app.db.session.commit()

    return json.dumps({"message": "ok"})


@api.route("/task/<task_id>/requeue")
@requires_auth
@with_task
def requeue_task(task):
    task_ids = requeue_tasks([task], current_app.db.session)
    if not task_ids:
        abort(400)
    current_app.db.session.commit()

    return json.dumps({"taskId": task_ids[0]})


@api.route("/queues")
@requires_auth
def get_queues():
    by_queue = {}

    q = (current_app.db.session.query(Task.status, Task.routing_key, func.count(Task.id))
         .filter(Task.status.in_(["QUEUED", "STARTED", "RETRY"]))
         .group_by(Task.status, Task.routing_key))

    for row in q:
        status, routing_key, count = row
        sub_dict = by_queue.setdefault(routing_key, {"QUEUED": 0, "STARTED": 0, "RETRY": 0, "ALL": 0,
                                                     "routing_key": routing_key})
        sub_dict[status] = count
        sub_dict["ALL"] += count

    return json.dumps({"result": list(by_queue.values())})


@api.route("/workers")
@requires_auth
def get_workers():
    workers = []
    for worker_name, stats in (current_app.celery_app.control.inspect().stats() or {}).items():
        workers.append({
            "name": worker_name,
            "broker": stats.get("broker", {}).get("transport"),
            "poolSize": stats.get("pool", {}).get("max-concurrency"),
            "tasks": stats.get("total", {}),
            "pid": stats.get("pid")
        })
    return json.dumps({"result": workers})


@api.route("/task", methods=["POST"])
@requires_auth
def create_task():
    data = request.get_json()
    task_id = current_app.celery_app.send_task(data["task"], kwargs=data["kwargs"], queue=data["queue"] or "celery")
    return json.dumps({"taskId": str(task_id)})


import os
from flask import Blueprint, send_from_directory

from ..auth import requires_auth

static = Blueprint('static', __name__, url_prefix='/static')


@static.route("/<folder>/<file>")
@requires_auth
def tasks(folder, file):
    folder_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), "frontend/dist/static", folder)

    return send_from_directory(folder_directory, file)



from contextlib import contextmanager
from logging import getLogger

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

MyResultModelBase = declarative_base()

SessionMaker = sessionmaker()

logger = getLogger(__name__)


@contextmanager
def session_ctx_manager():
    session = SessionMaker()
    try:
        yield session
    except:
        session.rollback()
        session.close()
        raise
    else:
        session.commit()
        session.close()


def prepare_models(engine):
    with engine.begin() as conn:
        pg_version = list(conn.execute("SELECT version();"))[0][0].split(" ")[1]
        pg_version_numbers = list(map(int, pg_version.split(".")))
        if pg_version_numbers[0] < 9 or (pg_version_numbers[0] == 9 and pg_version_numbers[1] < 5):
            logger.error("You should have a PostgreSQL version >= 9.5")
            return
        conn.execute("CREATE SCHEMA IF NOT EXISTS celery_jobs")
    MyResultModelBase.metadata.create_all(engine)

    # this happens pre-fork, so we need to dispose the engine on the main process as well
    engine.dispose()
    return True


class Task(MyResultModelBase):
    """Task result/status."""

    __tablename__ = 'tasks'
    __table_args__ = {'schema': 'celery_jobs'}

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    task_id = sa.Column(sa.Text, unique=True)
    status = sa.Column(sa.Text, index=True)
    name = sa.Column(sa.Text, default=None, index=True)
    routing_key = sa.Column(sa.String(50), default=None, index=True)
    result = sa.Column(sa.TEXT, nullable=True)
    args = sa.Column(sa.TEXT, nullable=True)
    kwargs = sa.Column(sa.TEXT, nullable=True)
    meta = sa.Column(JSONB, nullable=True)
    date_done = sa.Column(sa.DateTime(timezone=True), nullable=True, index=True)
    date_queued = sa.Column(sa.DateTime(timezone=True), nullable=True, index=True)
    date_started = sa.Column(sa.DateTime(timezone=True), nullable=True, index=True)
    eta = sa.Column(sa.DateTime(timezone=True), nullable=True, index=True)
    exception_type = sa.Column(sa.Text, default=None, index=True, nullable=True)
    traceback = sa.Column(sa.Text, nullable=True)

    def __init__(self, task_id):
        self.task_id = task_id

    def to_dict(self):
        return {
            'id': self.id,
            'task_id': self.task_id,
            'status': self.status,
            'result': self.result,
            'traceback': self.traceback,
            'date_done': self.date_done,
            'date_queued': self.date_queued,
            'date_started': self.date_started,
            'eta': self.eta,
            'name': self.name,
            'routing_key': self.routing_key,
            'exception_type': self.exception_type,
            'kwargs': self.kwargs,
            'args': self.args,
            'meta': self.meta
        }

    @property
    def serialized(self):
        return {
            'id': self.id,
            'task_id': self.task_id,
            'status': self.status,
            'result': self.result,
            'traceback': self.traceback,
            'date_queued': self.date_queued.isoformat() if self.date_queued else None,
            'date_done': self.date_done.isoformat() if self.date_done else None,
            'date_started': self.date_started.isoformat() if self.date_started else None,
            'eta': self.eta.isoformat() if self.eta else None,
            'name': self.name,
            'routing_key': self.routing_key,
            'exception_type': self.exception_type,
            'kwargs': self.kwargs,
            'args': self.args,
            'meta': self.meta
        }

    @classmethod
    def upsert(cls, task_id, on_conflict_update=None, **opts):
        if on_conflict_update is None:
            on_conflict_update = opts
        table = cls.__table__
        insert_stmt = insert(table).values(task_id=task_id, **opts)
        stm = insert_stmt.on_conflict_do_update(
            index_elements=[table.c.task_id],
            set_=on_conflict_update
        )
        with session_ctx_manager() as session:
            session.execute(stm)

    def __repr__(self):
        return '<Task {0.task_id} state: {0.status}>'.format(self)



from setuptools import setup, find_packages


def get_requirements():
    reqs = []
    for filename in ["requirements.txt"]:
        with open(filename, "r") as f:
            reqs += [x.strip() for x in f.readlines()]
    return reqs


setup(
    name='celery-dashboard',
    version='0.0.1',
    entry_points={
        'celery.commands': [
            'dashboard = celery_dashboard.command:CeleryDashboard',
        ],
    },
    install_requires=get_requirements(),
    packages=["celery_dashboard", "celery_dashboard.api"],
    include_package_data=True,
    description="A simple, extensible and full-featured dashboard for celery",
    author="Mehdi GMIRA",
    license='MIT',
    keywords=["celery", "python", "dashboard", "postgresql", "sql", "database"],
    platforms='any',
    python_requires='>=3.5'
)



import json

from celery import current_app

from .models import Task


def set_progress(task, progress):
    request = task.request
    Task.upsert(request.id, status="STARTED", name=task.name, args=dump(request.args), kwargs=dump(request.kwargs),
                routing_key=request.delivery_info["routing_key"], meta={"progress": progress})


def dump(data):
    try:
        return json.dumps(data)
    except TypeError:
        return repr(data)


def load(args, kwargs):
    try:
        if args:
            if args.startswith("(") and args.endswith(")"):
                args = "[" + args[1:-1] + "]"
            args = json.loads(args)
        if kwargs:
            kwargs = json.loads(kwargs)
    # args or kwargs were not jsonified, we do not requeue this kind of tasks
    # because that would imply pickling the arguments and it would be insecure
    except (ValueError, TypeError):
        return
    return args, kwargs


def cancel_tasks(tasks, session):
    to_rm = []
    to_set_as_cancelled = []
    count = 0
    for task in tasks:
        count += 1
        if task.status == "QUEUED":
            current_app.control.revoke(task.task_id)
            to_set_as_cancelled.append(task.task_id)
        else:
            to_rm.append(task.task_id)
        if len(to_rm) > 1000:
            session.query(Task).filter(Task.task_id.in_(to_rm)).delete(synchronize_session=False)
            to_rm = []
        if len(to_set_as_cancelled) > 1000:
            (session.query(Task).filter(Task.task_id.in_(to_set_as_cancelled))
             .update({'status': "CANCELLED"}, synchronize_session=False))
            to_set_as_cancelled = []
    if to_rm:
        session.query(Task).filter(Task.task_id.in_(to_rm)).delete(synchronize_session=False)
    if to_set_as_cancelled:
        (session.query(Task).filter(Task.task_id.in_(to_set_as_cancelled))
         .update({'status': "CANCELLED"}, synchronize_session=False))
    return count


def requeue_tasks(tasks, session):
    to_rm = []
    task_ids = []
    for task in tasks:
        args = task.args or []
        kwargs = task.kwargs or {}
        loaded_data = load(args, kwargs)
        if not loaded_data:
            continue
        to_rm.append(task.task_id)
        args, kwargs = loaded_data
        task_ids.append(current_app.send_task(task.name,
                                              args=args, kwargs=kwargs, queue=task.routing_key or "celery").task_id)
        if len(to_rm) > 1000:
            session.query(Task).filter(Task.task_id.in_(to_rm)).delete(synchronize_session=False)
            to_rm = []
    if to_rm:
        session.query(Task).filter(Task.task_id.in_(to_rm)).delete(synchronize_session=False)
    return task_ids
